import inspect
import pathlib
import sys

import astropy.units as u
import h5py
import healpy as hp
import numpy as np
from numba import njit
from tqdm import tqdm

from cosmoglobe.sky import components
from cosmoglobe.sky import Model
from cosmoglobe.utils import utils

# Model parameter group name as implemented in commander
param_group = 'parameters'
# These will be dropped from component lists
_ignored_comps = ['md', 'relquad']

#Current Cosmoglobe Sky Model as of BP9
COSMOGLOBE_COMPS = dict(
    ame=components.AME,
    cmb=components.CMB,
    dust=components.Dust,
    ff=components.FreeFree,
    radio=components.Radio,
    synch=components.Synchrotron,
)


def model_from_chain(file, nside=None, samples='all', burn_in=None, comps=None):
    """Returns a sky model from a commander3 chainfile.
    
    A cosmoglobe.sky.Model is initialized that represents the sky model used in 
    the given Commander run.

    Args:
    -----
    file (str):
        Path to commander3 hdf5 chainfile.
    nside (int):
        Healpix resolution parameter.
    sample (str, int, list):
        If sample is None, then the model will be initialized from sample 
        averaged maps. If sample is a string (or an int), the model will be 
        initialized from that specific sample. If sample is a list, then the 
        model will be initialized from an average over the samples in the list.
        Default: None
    burn_in (str, int):
        The sample number as a str or int where the chainfile is assumed to 
        have sufficently burned in. All samples before the burn_in are ignored 
        in the averaging process.
    comps (dict):
        Dictionary of which classes to use for each component. The keys must
        the comp group names in the chain file. If comps is None, a default set
        of components will be selected. Default: None

    Returns:
    --------
    (cosmoglobe.sky.Model):
        A sky model representing the results of a commander3 run.

    """
    model = Model(nside=nside)

    default_comps = COSMOGLOBE_COMPS

    if comps is None:
        comps = default_comps
    else:
        if not isinstance(comps, list):
            comps = [comps]
        comps = {comp:default_comps[comp] for comp in comps}

    if not comps:
        raise ValueError('No comps selected')

    chain_components = _get_components(file)
    component_list = [comp for comp in comps if comp in chain_components]

    if samples == 'all':
        samples = _get_samples(file)
    elif samples == -1:
        samples = _get_samples(file)[-1]
    elif isinstance(samples, int):
        samples = _int_to_sample(samples)
    if burn_in is not None:
        if len(samples) > burn_in:
            samples = samples[burn_in:]
        else:
            raise ValueError('burn_in sample is out of range')

    print('Loading components from chain')
    with tqdm(total=len(component_list), file=sys.stdout) as pbar:
        padding = len(max(component_list, key=len))
        for comp in component_list:
            pbar.set_description(f'{comp:<{padding}}')
            comp = comp_from_chain(file, comp, comps[comp], 
                                     nside, samples)                                           
            model._insert_component(comp)
            pbar.update(1)

        pbar.set_description('Done')

    return model


def comp_from_chain(file, component, component_class, model_nside, samples):
    """Returns a sky component from a commander3 chainfile.
    
    A sky component that subclasses cosmoglobe.sky.Component is initialized 
    from a given commander run.

    Args:
    -----
    file (str):
        Path to commander3 hdf5 chainfile.
    component (str):
        Name of a sky component. Must match the hdf5 component group name.
    sample (str, int, list):
        If sample is 'mean', then the model will be initialized from sample 
        averaged maps. If sample is a string (or an int), the model will be 
        initialized from that specific sample. If sample is a list, then the 
        model will be initialized from an average over the samples in the list.
        Default: 'mean'
    burn_in (str, int):
        The sample number as a str or int where the chainfile is assumed to 
        have sufficently burned in. All samples before the burn_in are ignored 
        in the averaging process.

    Returns:
    --------
    (sub class of cosmoglobe.sky.Component):
        A sky component initialized from a chain.

    """
    # Getting component parameters from chain
    parameters = _get_component_params(file, component)
    freq_ref = (parameters['nu_ref']*u.Hz).to(u.GHz)
    fwhm_ref = (parameters['fwhm']*u.arcmin).to(u.rad)
    nside = parameters['nside']
    if parameters['polarization'] == 'True':
        comp_is_polarized = True
    else:
        comp_is_polarized = False

    # Commander outputs units in uK_RJ for all comps except for CMB which is in
    # K_CMB. This is manually handeled in the CMB component. NB! If this 
    # changes in future Commander versions, this part needs to be updated.
    amp_unit = u.uK

    # Getting arguments required to initialize component
    args_list = _get_comp_args(component_class) 
    print(args_list)
    args = {}
    if isinstance(samples, list) and len(samples) > 1:
        get_items = _get_averaged_items
    else:
        get_items = _get_items

    # Find which args are alms and which are precomputed maps
    alm_names = []
    map_names = []
    other_items_names = []
    for arg in args_list:
        if arg != 'freq_ref':
            if _item_alm_exists(file, component, arg):
                alm_names.append(arg)
            elif _item_map_exists(file,component, arg):
                map_names.append(arg)
            elif _item_exists(file, component, arg):
                other_items_names.append(arg)
            elif arg == 'nside':
                args['nside'] = nside
            else:
                raise KeyError(f'item {arg} is not present in the chain')


    other_items_ = get_items(
        file, samples, component, [item for item in other_items_names]
    )
    other_items = dict(zip(other_items_names, other_items_))
    maps_ = get_items(
        file, samples, component, [f'{map_}_map' for map_ in map_names]
    )
    maps = dict(zip(map_names, maps_))
    if maps:
        if model_nside is not None and nside != model_nside:
            maps = {key:hp.ud_grade(value, model_nside) 
                    if isinstance(value, np.ndarray) 
                    else value for key, value in maps.items()}
    args.update(maps)
    args.update(other_items)
    if model_nside is None:
        model_nside = nside

    alms_ = get_items(
        file, samples, component, [f'{alm}_alm' for alm in alm_names]
    )
    alms = dict(zip(alm_names, alms_))
    alms_lmax_ = _get_items(
        file, samples[-1], component, [f'{alm}_lmax' for alm in alm_names]
    )
    alms_lmax = dict(zip(alm_names, [int(lmax) for lmax in alms_lmax_]))

    for key, value in alms.items():
        unpacked_alm = unpack_alms_from_chain(value, alms_lmax[key])
        if key == 'amp' and value.shape[0] == 3:
            pol = True
        else:
            pol = False
                
        alms[key] = hp.alm2map(unpacked_alm, 
                          nside=model_nside, 
                          lmax=alms_lmax[key], 
                          fwhm=fwhm_ref.value,
                          pol=pol,
                          ).astype('float32')

    args.update(alms)
    args['amp'] *= amp_unit
    args = utils.set_spectral_units(args)
    scalars = utils.extract_scalars(args) # dont save scalar maps
    args.update(scalars)
    if 'freq_ref' in args_list:
        if comp_is_polarized:
            freq = u.Quantity(freq_ref[:-1])
        else:
            freq = u.Quantity(freq_ref[0])
        args['freq_ref'] = freq

    return component_class(**args)


def _get_comp_args(component_class):
    """Returns a list of arguments needed to initialize a component"""
    ignored_args = ['self']
    arguments = inspect.getargspec(component_class.__init__).args
    return [arg for arg in arguments if arg not in ignored_args]


def _get_samples(file):
    """Returns a list of all samples present in a chain file"""
    with h5py.File(file, 'r') as f:
        samples = list(f.keys())

    try:
        samples.remove(param_group)
    except:
        print("Warning: Using an old h5 commander format without param_group")

    return samples


def _sample_to_int(samples, start=0):
    """Converts a sample or a list of samples to integers"""
    if isinstance(samples, (list, tuple)):
        return [int(sample) + start for sample in samples]
    
    return int(samples) + start


def _int_to_sample(samples, start=0):
    """Converts an integer or multiple integers to sample string format"""
    if isinstance(samples, (list, tuple)):
        return [f'{sample + start:06d}' for sample in samples]
    
    return f'{samples + start:06d}'


def _get_components(file, ignore_comps=True):
    """Returns a list of all components present in a chain file"""
    with h5py.File(file, 'r') as f:
        components = list(f[f'{1:06d}'].keys())
    if ignore_comps:    
        return [comp for comp in components if comp not in _ignored_comps]

    return components


def _get_component_params(file, component):
    """Returns a dictionary of the model parameters of a component"""
    return_params = {}
    with h5py.File(file, 'r') as f:
        params = f[param_group][component]
        for param, value in params.items():
            if isinstance(value[()], bytes):
                return_params[param] = value.asstr()[()]
            else:
                return_params[param] = value[()]

        return return_params
    

def _get_items(file, sample, component, items):
    """Return the value of one or many items for a component in the chain file.

    Args:
    -----
    file: str
        Path to h5 file.
    sample : str
        sample name.
    component: str
        Component group name.
    items: str, list, tuple
        Name of item to extract, or a list of names.

    Returns:
    --------
    list
        List of items extracted from the chain file.
    
    """
    with h5py.File(file, 'r') as f:
        items_to_return = []
        try:
            for item in items:
                items_to_return.append(f[sample][component][item][()])

            return items_to_return
        except TypeError:
            return f[sample][component][items][()]


def _get_averaged_items(file, samples, component, items):
    """Return the averaged value of one or many item for a component in the 
    chain file.

    Args:
    -----
    file: str
        Path to h5 file.
    samples : list
        List of samples to average over.
    component:
        Component group.
    items: str, list, tuple
        Name of item to extract, or a list of names. Items must be of types
        compatible with integer division.

    Returns:
    --------
    list
        List of items averaged over samples from the chain file.
    
    """
    if not items:
        return []

    with h5py.File(file, 'r') as f:
        if isinstance(items, (tuple, list)):
            items_to_return = []
            for sample in samples:
                
                for idx, item in enumerate(items):
                    try:
                        items_to_return[idx] += (
                            f[sample][component][item][()]
                        )
                    except IndexError:
                        items_to_return.append(
                            f[sample][component][item][()]
                        )

            return [item/len(samples) for item in items_to_return]

        for sample in samples:
            try:
                item_to_return += f[sample][component][items][()]
            except UnboundLocalError:
                item_to_return = f[sample][component][items][()]

        return item_to_return/len(samples)


def _item_alm_exists(file, component, item):
    """Returns True if component contains alms for the given item, else 
    returns False.
    """
    sample = _get_samples(file)[-1]

    with h5py.File(file, 'r') as f:
        params = list(f[sample][component].keys())

    if f'{item}_alm' in params:
        return True

    return False


def _item_map_exists(file, component, item):
    """Returns True if component contains precomputed map for item, else 
    returns False.
    """
    sample = _get_samples(file)[-1]

    with h5py.File(file, 'r') as f:
        params = list(f[sample][component].keys())

    if f'{item}_map' in params:
        return True

    return False

def _item_exists(file, component, item):
    """Returns True if component contains the item (array or scalar), else 
    returns False.
    """
    sample = _get_samples(file)[-1]

    with h5py.File(file, 'r') as f:
        params = list(f[sample][component].keys())

    if f'{item}' in params:
        return True

    return False


@njit
def unpack_alms_from_chain(data, lmax):
    """Unpacks alms from the Commander chain output.

    Unpacking algorithm: 
    https://github.com/trygvels/c3pp/blob/2a2937926c260cbce15e6d6d6e0e9d23b0be1262/src/tools.py#L9

    TODO: look over this function and see if it can be improved.

    Args:
    -----
    data (np.ndarray)
        alms from a commander chainfile.
    lmax : int
        Maximum value for l used in the alms.

    Returns:
    --------
    alms (np.ndarray)
        Unpacked version of the Commander alms (2-dimensional array)
    
    """
    n = len(data)
    n_alms = int(lmax * (2*lmax+1 - lmax) / 2 + lmax+1)
    alms = np.zeros((n, n_alms), dtype=np.complex128)

    for sigma in range(n):
        i = 0
        for l in range(lmax+1):
            j_real = l**2 + l
            alms[sigma, i] = complex(data[sigma, j_real], 0.0)
            i += 1

        for m in range(1, lmax+1):
            for l in range(m, lmax+1):
                j_real = l**2 + l + m
                j_comp = l**2 + l - m
                alms[sigma, i] = complex(data[sigma, j_real], 
                                         data[sigma, j_comp],)/ np.sqrt(2.0)
                i += 1

    return alms


def chain_to_h5(chainfile, output_dir, nside=None, burn_in=None):
    """Outputs the contents of a chainfile to model hdf5 files for a set of 
    resulutions. 

    Parameters
    ----------
    chainfile : str, `pathlib.PosixPath`
        Path to commander3 hdf5 chainfile.
    output_dir : str, `pathlib.PosixPath`
        Path to where model hdf5 files will be output.
    nside : int
        Custom nside value. Overrides the DEFAULT_NSIDES. Default: None.
    burn_in : Sample at which the chain has "burned in".
    """

    DEFAULT_NSIDES = [
        2**res for res in range(12)     # [1, 2, ... , 2048]
    ]

    NSIDES = DEFAULT_NSIDES if nside is None else nside
    for nside in NSIDES:
        model = model_from_chain(chainfile, nside, burn_in=burn_in)
        model_to_h5(model, output_dir)


def model_to_h5(model, output_dir):
    """Outputs a `cosmoglobe.sky.Model` to a hdf5 model file.

    Parameters
    ----------
    model : `cosmoglobe.sky.Model`
        A cosmoglobe sky model.
    output_dir : str, `pathlib.PosixPath`
        Path to where model hdf5 files will be output.
    """

    dirname = pathlib.Path(output_dir)
    dirname.mkdir(parents=True, exist_ok=True)
    filename = dirname / f'model_{model.nside}.h5'

    with h5py.File(filename, 'w') as f:
        for comp in model:
            grp = f.create_group(comp.label)
            amp = grp.create_dataset('amp', data=comp.amp.value)
            amp.attrs['unit'] = comp.amp.unit.to_string()
            if comp.freq_ref is not None:
                freq = grp.create_dataset('freq_ref', data=comp.freq_ref.value)
                freq.attrs['unit'] = comp.freq_ref.unit.to_string()

            sp_grp = grp.create_group('spectral_parameters')
            for key, value in comp.spectral_parameters.items():
                if isinstance(value, u.Quantity):
                    dset = sp_grp.create_dataset(key, data=value.value)
                    dset.attrs['unit'] = value.unit.to_string()
                else:
                    sp_grp.create_dataset(key, data=value)



def model_from_h5(filename):
    """Initializes a `cosmoglobe.sky.Model` from a hdf5 file with a specific
    format.

    Parameters
    ----------
    filename : str, `pathlib.PosixPath`
        Filename of the hdf5 file from which to read in model.
    """

    filename = pathlib.Path(filename)
    model = Model()
    nside = None
    with h5py.File(filename, 'r') as f:
        for comp in f:
            amp_dset = f.get(f'{comp}/amp')
            amp = u.Quantity(
                value=amp_dset[()], 
                unit=amp_dset.attrs.get('unit', None)
            )
            freq_dset = f.get(f'{comp}/freq_ref')
            if freq_dset:
                freq_ref = u.Quantity(
                    value=freq_dset[()], 
                    unit=freq_dset.attrs.get('unit', None)
                ) 
            else:
                freq_ref = None
            
            spectral_parameters = {}
            for spec in f[comp]['spectral_parameters']:
                dset = f.get(f'{comp}/spectral_parameters/{spec}')
                spectral_parameters[spec] = u.Quantity(
                    value=dset[()],
                    unit=dset.attrs.get('unit', None)
                )

            if nside is None:
                nside = hp.get_nside(amp)

            component = COSMOGLOBE_COMPS[comp]
            if comp == 'radio':
                model._add_component_to_model(
                    component(
                        amp=amp, 
                        freq_ref=freq_ref, 
                        nside=nside,
                        **spectral_parameters
                    )
                )            
            else:
                model._add_component_to_model(
                    component(amp=amp, freq_ref=freq_ref, **spectral_parameters)
                )

    return model


from cosmoglobe.sky.model import Model
from cosmoglobe.sky.components import (
    AME,
    CMB,
    Dust,
    FreeFree,
    Radio,
    Synchrotron,
)
from cosmoglobe.hub import sky_model
from cosmoglobe.chain.h5 import model_from_chain
from cosmoglobe.plot import plot

__all__ = ['Class']



import os
import healpy as hp
import numpy as np
import matplotlib.pyplot as plt

from functools import partial
from .plottools import load_cmap, fmt, autoparams, apply_logscale, set_style, make_fig


def gnomplot(
    map_,
    lon,
    lat,
    auto=None,
    sig=0,
    size=20,
    vmin=None,
    vmax=None,
    rng=None,
    title=None,
    ltitle=None,
    unit=None,
    cmap="planck",
    graticule=False,
    logscale=False,
    colorbar=True,
    fwhm=0.0,
    remove_dip=False,
    remove_mono=False,
    fontsize=11,
    figsize=(5, 6),
    darkmode=False,
    fignum=None,
    subplot=None,
    hold=False,
    reuse_axes=False,
):
    """
    Gnomonic view plotting.
    """
    # Set plotting rcParams
    set_style(darkmode)

    xsize = 5000
    reso = size * 60 / xsize
    if isinstance(map_, str):
        x = hp.read_map(map_, field=sig, verbose=False, dtype=None)
    else:
        if map_.ndim > 1:
            x = map_[sig]
        else:
            x = map_

    nside = hp.get_nside(x)

    if float(fwhm) > 0:
        x = hp.smoothing(x, fwhm=fwhm * (2 * np.pi) / 21600, lmax=3 * nside,)
    if remove_dip:
        x = hp.remove_dipole(x, gal_cut=30, copy=True, verbose=True)
    if remove_mono:
        x = hp.remove_monopole(x, gal_cut=30, copy=True, verbose=True)

    proj = hp.projector.GnomonicProj(
        rot=[lon, lat, 0.0], coord="G", xsize=xsize, ysize=xsize, reso=reso
    )
    reproj_im = proj.projmap(x, vec2pix_func=partial(hp.vec2pix, nside))

    if rng is not None:
        vmin = -rng
        vmax = rng

    params = autoparams(auto, sig, title, ltitle, unit, None, logscale, cmap)
    vmin, vmax = (params["ticks"][0], params["ticks"][-1])
    if params["logscale"]:
        x, (vmin, vmax) = apply_logscale(x, [vmin, vmax], linthresh=1)

    cmap = load_cmap(params["cmap"], params["logscale"])

    # Format for latex
    for i in ["title", "unit", "left_title"]:
        if params[i] and params[i] != "":
            params[i] = r"$" + params[i] + "$"

    fig, ax = make_fig(figsize, fignum, hold, subplot, reuse_axes,)
    image = plt.imshow(
        reproj_im,
        origin="lower",
        interpolation="nearest",
        vmin=vmin,
        vmax=vmax,
        cmap=cmap,
    )
    plt.xticks([])
    plt.yticks([])
    plt.text(
        0.95,
        0.95,
        params["title"],
        color="black",
        va="top",
        ha="right",
        transform=ax.transAxes,
        fontsize=fontsize + 4,
    )
    plt.text(
        0.05,
        0.95,
        params["left_title"],
        color="black",
        va="top",
        transform=ax.transAxes,
        fontsize=fontsize + 4,
    )
    if colorbar:
        # colorbar
        from matplotlib.ticker import FuncFormatter

        cb = plt.colorbar(
            image,
            orientation="horizontal",
            shrink=0.5,
            pad=0.03,
            format=FuncFormatter(fmt),
        )
        cb.ax.tick_params(which="both", axis="x", direction="in", labelsize=fontsize)
        cb.ax.xaxis.set_label_text(params["unit"])
        cb.ax.xaxis.label.set_size(fontsize)

    if graticule:
        hp.graticule()



from typing import List

import astropy.units as u
import healpy as hp
import numpy as np

from cosmoglobe.sky.base import (
    _Component, 
    _DiffuseComponent, 
    _PointSourceComponent,
)
from cosmoglobe.utils import utils
from cosmoglobe.utils.utils import CompState, ModelError


class Model:
    r"""Sky model object representing the Cosmoglobe Sky Model.

    This class acts as a container for the various components making up 
    the Cosmoglobe Sky Model, and provides methods to simulate the sky. 
    The primary use case of this class is to call its ``__call__`` 
    method, which simulates the sky at a single frequency :math:`\nu`, 
    or integrated over a bandpass :math:`\tau`.
    
    Attributes
    ----------
    AME : `cosmoglobe.sky.components.AME`
        The AME sky component.
    CMB : `cosmoglobe.sky.components.CMB`
        The CMB sky component.
    dust : `cosmoglobe.sky.components.Dust`
        The dust sky component.
    ff : `cosmoglobe.sky.components.FreeFree`
        The free-free sky component.
    radio : `cosmoglobe.sky.components.Radio`
        The radio sky component.
    synch : `cosmoglobe.sky.components.Synchrotron`
        The synchrotron sky component.
    nside : int
        Healpix resolution of the maps in sky model.

    Methods
    -------
    __call__
    disable
    enable

    See Also
    --------
    __call__ : 
        Simulates the sky at a given frequency :math:`\nu` or over a 
        bandpass :math:`\tau` given the Cosmoglobe Sky Model.

    Examples
    --------
    Inspecting the model:

    >>> model = skymodel(nside=256)
    >>> print(model)
    Model(
      nside: 256
      components( 
        (ame): AME(nu_p)
        (cmb): CMB()
        (dust): Dust(beta, T)
        (ff): FreeFree(Te)
        (radio): Radio(specind)
        (synch): Synchrotron(beta)
      )
    )

    Simulating the full sky emission at some frequency, given a beam 
    FWHM:

    >>> import astropy.units as u
    >>> emission = model(50*u.GHz, fwhm=30*u.arcmin)
    >>> print(emission)
    Smoothing point sources...
    Smoothing diffuse emission...
    [[ 2.25809786e+03  2.24380103e+03  2.25659060e+03 ... -2.34783682e+03
      -2.30464421e+03 -2.30387946e+03]
     [-1.64627550e+00  2.93583564e-01 -1.06788937e+00 ... -1.64354585e+01
       1.60621841e+01 -1.05506092e+01]
     [-4.15682825e+00  3.08881971e-01 -1.02012415e+00 ...  5.44745701e+00
      -4.71776995e+00  4.39850830e+00]] uK
    """

    nside: int = None 
    components: List[_Component] = None, 

    def __init__(self, nside=None, components=None):
        """Initializing a sky model. 

        Parameters
        ----------
        nside : int, optional
            Healpix resolution of the maps in sky model (the default is 
            None, in which the model automatically detects the nside from
            the components).
        components : list, optional
            A list of `cosmoglobe.sky.base._Component` objects that 
            constitutes the sky model (by default this is None and the 
            components are iteratively added as they are read from a 
            commander3 chain).
        """

        self.nside = nside
        self._components = {}

        if components is not None:
            for component in components:
                self._add_component_to_model(component)

    def _add_component_to_model(self, component):
        """Adds a new component to the model.
        
        Parameters
        ----------
        component : `cosmoglobe.sky.base._Component`
            Component to be added to the sky model.
        """

        if not isinstance(component, _Component):
            raise TypeError(
                f'component must be a subclass of {_Component}'
            )
        name = component.label
        if name in self._components:
            raise KeyError(
                f'component {name} is already a part of the model'
            )

        setattr(self, name, component)
        self._components[name] = [component, CompState.ENABLED]

        if self.nside is None:
            self.nside = hp.get_nside(component.amp)

    @u.quantity_input(
        freqs=u.Hz, 
        bandpass=(u.Jy/u.sr, u.K, None), 
        fwhm=(u.rad, u.deg, u.arcmin)
    )
    def __call__(
        self, freqs, bandpass=None, fwhm=0.0 * u.rad, output_unit=u.uK
    ):
        r"""Simulates the full model sky emission. 

        This method computes the full model sky emission (sum of all 
        component emission) for a single frequency :math:`\nu` or 
        integrated over a  bandpass :math:`\tau`.

        Parameters
        ----------
        freqs : `astropy.units.Quantity`
            A frequency, or a list of frequencies for which to evaluate the
            sky emission.
        bandpass : `astropy.units.Quantity`, optional
            Bandpass profile corresponding to the frequencies. Default is 
            None. If `bandpass` is None and `freqs` is a single frequency,
            a delta peak is assumed (unless `freqs` is a list of 
            frequencies, for which a top-hat bandpass is used to perform  
            bandpass integration instead).
        fwhm : `astropy.units.Quantity`, optional
            The full width half max parameter of the Gaussian (Default is 
            0.0, which indicates no smoothing of output maps).
        output_unit : str, `astropy.units.UnitBase`, optional
            The desired output units of the emission. The supported units are
            :math:`\mathrm{\mu K_{RJ}}` and :math:`\mathrm{MJ/sr}` (By 
            default the output unit of the model is always in 
            :math:`\mathrm{\mu K_{RJ}}`. 

        Returns
        -------
        `astropy.units.Quantity`
            The full model emission.

        Notes
        -----
        This function computes the following expression (assuming that 
        all default Cosmoglobe Sky components are present in the model):

        .. math::

            \boldsymbol{s}_{\mathrm{RJ}}(\nu) &=\boldsymbol{a}_{\mathrm{CMB}}
             \frac{x^{2} \mathrm{e}^{x}}{\left(\mathrm{e}^{x}-1\right)^{2}} 
            \frac{\left(\mathrm{e}^{x_{0}}-1\right)^{2}}{x_{0}^{2} 
            \mathrm{e}^{x_{0}}}\\
            &+\boldsymbol{a}_{\mathrm{s}}\left(\frac{\nu}{\nu_{0, 
            \mathrm{~s}}}\right)^{\beta_{\mathrm{s}}}\\
            &+\boldsymbol{a}_{\mathrm{ff}} \frac{g_{\mathrm{ff}}
            \left(\nu ; T_{e}\right)}{g_{\mathrm{ff}}\left(\nu_{0, 
            \mathrm{ff}} ; T_{e}\right)}\left(\frac{\nu_{0, 
            \mathrm{ff}}}{\nu}\right)^{2}\\
            &+\boldsymbol{a}_{\mathrm{sd}}\left(\frac{\nu_{0, 
            \mathrm{sd}}}{\nu}\right)^{2} \frac{s_{0}^{\mathrm{sd}}
            \left(\nu \cdot \frac{\nu_{p}}{30.0\; \mathrm{GHz}}\right)}
            {s_{0}^{\mathrm{sd}}\left(\nu_{0, \mathrm{sd}} \cdot 
            \frac{\nu_{p}}{30.0\; \mathrm{GHz}}\right)} \\
            &+\boldsymbol{a}_{\mathrm{d}}\left(\frac{\nu}{\nu_{0, 
            \mathrm{~d}}}\right)^{\beta_{\mathrm{d}}+1} 
            \frac{\mathrm{e}^{h \nu_{0, \mathrm{~d}} 
            / k T_{\mathrm{d}}}-1}{\mathrm{e}^{\mathrm{h} \nu 
            / k T_{\mathrm{d}}}-1}+\\&+\sum_{j=1}^{N_{\mathrm{src}}} 
            \boldsymbol{a}_{\mathrm{src}}^{j}\left(\frac{\nu}{\nu_{0, 
            \mathrm{src}}}\right)^{\alpha_{j, \mathrm{src}}-2}.

        For more information on the current implementation of the 
        Cosmoglobe Sky Model, see `BeyondPlanck (2020), Section 3.6 
        <https://arxiv.org/pdf/2011.05609.pdf>`_.

        Examples
        --------
        Simulated full sky emission at :math:`50\; \mathrm{GHz}`:

        >>> from cosmoglobe import skymodel
        >>> import astropy.units as u
        >>> model = skymodel(nside=256) 
        >>> model(50*u.GHz)[0]
        [ 2234.74893115  2291.99921295  2323.98779311 ... -2320.74732945
         -2271.54465982 -2292.22248419] uK

        Simulated full sky emission at :math:`500\; \mathrm{GHz}` 
        smoothed with a :math:`50\; '` Gaussian beam, outputed in units of 
        :math:`\mathrm{MJy} / \mathrm{sr}`:

        >>> model(500*u.GHz, fwhm=50*u.arcmin, output_unit='MJy/sr')[0]
        [ 0.267749    0.26991688  0.28053964 ... -0.15846278 -0.15269807
         -0.14408377] MJy / sr
        """

        shape = (3, hp.nside2npix(self.nside))
        diffuse_emission = np.zeros(shape)
        point_source_emission = np.zeros(shape)

        # The output unit may be a string denoting for instance K_CMB, which
        # is critical information for the following routines. However, we 
        # need to initialize the emission arrays with astropy units.
        _output_unit = utils.str_to_astropy_unit(output_unit)
        diffuse_emission = u.Quantity(diffuse_emission, unit=_output_unit)
        point_source_emission = u.Quantity(
            point_source_emission, unit=_output_unit
        )

        for comp in self:
            if isinstance(comp, _DiffuseComponent):
                comp_emission = comp(freqs, bandpass, output_unit=output_unit)
                for idx, row in enumerate(comp_emission):
                    diffuse_emission[idx] += row

            elif isinstance(comp, _PointSourceComponent):
                comp_emission = comp(
                    freqs, bandpass, fwhm=fwhm, output_unit=output_unit
                )
                for idx, row in enumerate(comp_emission):
                    point_source_emission[idx] += row

        if fwhm != 0.0 and diffuse_emission.value.any():
            # Instead of calling the diffuse components _smooth function
            # it is more efficient to sum all the emission first and then
            # only perform a singled smoothing operation.
            print('Smoothing diffuse emission...')
            diffuse_emission = u.Quantity(
                hp.smoothing(diffuse_emission, fwhm.to(u.rad).value),
                unit=diffuse_emission.unit
            )

        return diffuse_emission + point_source_emission

    def disable(self, comp_label):
        """Disable a component in the model.

        Parameters
        ----------
        comp_label : str
            The component label.

        Raises
        ------
        KeyError
            If the component is not currently enabled in the model.
        """

        try:
            if self._components[comp_label][1] is CompState.ENABLED:
                self._components[comp_label][1] = CompState.DISABLED
            else:
                raise ModelError(f'{comp_label!r} is already disabled')
        except KeyError:
            raise KeyError(f'{comp_label!r} is not a component in the model')

    def enable(self, comp_label):
        """Enable a disabled component.

        Parameters
        ----------
        comp_label : str
            The component label.
        
        Raises
        ------
        KeyError
            If the component is not currently disabled in the model.
        """

        try:
            if self._components[comp_label][1] is CompState.DISABLED:
                self._components[comp_label][1] = CompState.ENABLED
            else:
                raise ModelError(f'{comp_label!r} is already enabled')
        except KeyError:
            raise KeyError(f'{comp_label!r} is not a component in the model')

    @property
    def components(self):
        """Returns a list of enabled components."""
        return [
            comp[0] for comp in self._components.values()
            if comp[1] is CompState.ENABLED
        ]

    def __iter__(self):
        """Creates an iterator from the components list such that the 
        model can be iterated over.
        """       

        return iter(self.components)

    def __repr__(self):        
        """Representation of the Model and all enabled components."""

        reprs = []
        for component in self.components:
            component_repr = repr(component) + '\n'
            reprs.append(f'({component.label}): {component_repr}')

        main_repr = f'Model('
        main_repr += f'\n  nside: {self.nside}'
        main_repr += '\n  components( '
        main_repr += '\n    ' + '    '.join(reprs)
        main_repr += f'  )'
        main_repr += f'\n)'

        return main_repr


import numpy as np
import astropy.units as u

import cosmoglobe.utils.constants as const

u.quantity_input(freq=u.Hz, T=u.K)
def blackbody_emission(freq, T):
    """Returns the blackbody emission.
    
    Computes the emission emitted by a blackbody with with temperature 
    T at a frequency freq in SI units [W / m^2 Hz sr].

    Parameters
    ----------
    freq : `astropy.units.Quantity`
        Frequency [Hz].
    T : `astropy.units.Quantity`:
        Temperature of the blackbody [K]. 

    Returns
    -------
    `astropy.units.Quantity`:
        Blackbody emission [W / m^2 Hz sr].
    """

    # Avoiding overflow and underflow.
    T = T.astype(np.float64)

    term1 = (2*const.h*freq**3) / const.c**2
    term2 = np.expm1((const.h*freq) / (const.k_B*T))

    emission = term1 / term2 / u.sr

    return emission.to('W / m^2 Hz sr')


u.quantity_input(freq=u.Hz, T_e=u.K)
def gaunt_factor(freq, T_e):
    """Returns the Gaunt factor.
    
    Computes the gaunt factor for a given frequency and electron 
    temperaturein SI units.

    Parameters
    ----------
    freq : `astropy.units.Quantity`
        Frequency [Hz].   
    T_e : astropy.units.Quantity`
        Electron temperature [K].

    Returns
    -------
    gaunt_factor : `astropy.units.Quantity`
        Gaunt Factor.
    """
    
    # Avoiding overflow and underflow.
    T_e = T_e.astype(np.float64)
    T_e = (T_e.to(u.kK)).value / 10
    freq = (freq.to(u.GHz)).value   
 
    gaunt_factor = np.log(
        np.exp(5.96 - (np.sqrt(3)/np.pi) * np.log(freq*T_e**-1.5)) + np.e
    )

    return u.Quantity(gaunt_factor, unit=u.dimensionless_unscaled)


u.quantity_input(freq=u.Hz, T=u.K)
def thermodynamical_to_brightness(freq, T=const.T_0):
    """Conversion factor between K_CMB and K_RJ.

    Returns the conversion factor between thermodynamical and brightness 
    temperatures [K_CMB -> K_RJ].

    Parmaeters
    ----------
    freq : `astropy.units.Quantity`
        Frequency [K]. 

    Returns:
    --------
    `astropy.units.Quantity`
        Conversion factor.
    """  

    x = (const.h*freq) / (const.k_B*T)
    return ((x**2 * np.exp(x)) / (np.expm1(x)**2)).si


u.quantity_input(freq=u.Hz, T=u.K)
def brightness_to_thermodynamical(freq, T=const.T_0):
    """Conversion factor between K_RJ and K_CMB.

    Returns the conversion factor between thermodynamical and brightness 
    temperatures [K_RJ -> K_CMB].

    Parmaeters
    ----------
    freq : `astropy.units.Quantity`
        Frequency [K]. 

    Returns:
    --------
    `astropy.units.Quantity`
        Conversion factor.
    """  

    return 1 / thermodynamical_to_brightness(freq, T)


import astropy.units as u
import numpy as np
from scipy.interpolate import RectBivariateSpline

import cosmoglobe.utils.constants as const

@u.quantity_input(bandpass=(u.Jy/u.sr, u.K), freqs=u.Hz)
def get_normalized_bandpass(bandpass, freqs, unit=u.K):
    """Returns a bandpass profile normalized to unity when integrated over all 
    frequencies corresponding to the profile in units of K_RJ.

    Parameters
    ----------
    bandpass: `astropy.units.Quantity`
        Bandpass profile in signal units.    
    freqs: `astropy.units.Quantity`
        Frequencies corresponding to the bandpass profile.
    unit: `astropy.units.UnitBase`
        The unit we want to convert to.

    Returns
    -------
    `astropy.units.Quantity`
        Normalized bandpass
    """

    bandpass = bandpass.to(
        unit=unit, equivalencies=u.brightness_temperature(freqs)
    )

    return bandpass/np.trapz(bandpass, freqs)


@u.quantity_input(bandpass=u.Hz**-1, freqs=u.Hz)
def get_bandpass_coefficient(bandpass, freqs, output_unit):
    """Returns the bandpass coefficient.
    
    Computes the bandpass coefficent for a map that has been calibrated in 
    K_RJ to some output unit, after taking into account the bandpass.

    Parameters
    ----------
    bandpass: `astropy.units.Quantity`
        Normalized bandpass profile [Hz^-1].    
    freqs: `astropy.units.Quantity`
        Frequencies corresponding to the bandpass profile.
    unit: `astropy.units.UnitBase`
        The unit we want to convert to.

    Returns
    -------
    bandpass_coefficient: `astropy.units.Quantity`
        The bandpass coefficient.
    """

    intensity_derivative_i = b_rj(freqs)

    #Selecting the intensity derivative expression given an output_unit
    try:
        output_unit = u.Unit(output_unit)
        if u.K in output_unit.si.bases:
            intensity_derivative_j = b_rj(freqs)
        elif output_unit == u.MJy/u.sr:
            intensity_derivative_j = b_iras(freqs, np.mean(freqs))

    except ValueError:
        if isinstance(output_unit, str):
            if output_unit.lower().endswith('k_rj'):
                intensity_derivative_j = b_rj(freqs)
            elif output_unit.lower().endswith('k_cmb'):
                intensity_derivative_j = b_cmb(freqs)
            else:
                raise NotImplementedError(
                    f'output_unit {output_unit} is not implemented'
                )
    bandpass_coefficient = (
        np.trapz(bandpass*intensity_derivative_i, freqs)/
        np.trapz(bandpass*intensity_derivative_j, freqs)
    )

    return bandpass_coefficient


def get_bandpass_scaling(freqs, bandpass, comp):
    """Returns the frequency scaling factor given a bandpass profile and a
    corresponding frequency array. 
    
    Parameters
    ----------
    freqs : `astropy.units.Quantity`
        List of frequencies.
    bandpass : `astropy.units.Quantity`
        Normalized bandpass profile.
    Returns
    -------
    bandpass_scaling : float, `numpy.ndarray`
        Frequency scaling factor given a bandpass.
    """
    
    grid = get_interpolation_grid(comp.spectral_parameters)
    if not grid:
    # Component does not have any spatially varying spectral parameters.
    # In this case we simply integrate the emission at each frequency 
    # weighted by the bandpass.
        freq_scaling = comp._get_freq_scaling(
            freqs, comp.freq_ref, **comp.spectral_parameters
        )
        return np.expand_dims(
            np.trapz(freq_scaling*bandpass, freqs), axis=1
        )
    elif len(grid) == 1:
    # Component has one sptatially varying spectral parameter. In this 
    # scenario we perform a 1D-interpolation in spectral parameter space.
        return interp1d(freqs, bandpass, grid, comp)
    elif len(grid) == 2:    
    # Component has two sptatially varying spectral parameter. In this 
    # scenario we perform a 2D-interpolation in spectral parameter space.
        return interp2d(freqs, bandpass, grid, comp)
    else:
        raise NotImplementedError(
            'Bandpass integration for comps with more than two spectral '
            'parameters is not currently supported'
        )


def get_interpolation_grid(spectral_parameters):
    """Returns a interpolation range.
    
    Computes the interpolation range of the spectral parameters of a 
    sky component. We use a regular grid with n points for the range.
    
    Parameters
    ----------
    spectral_parameters: dict
        Dictionary containing the spectral parameters of a given component.

    Returns
    -------
    interp_parameters: dict
        Dictionary with a interpolation grid for each spatially varying
        spectral parameter.
    """

    dim = 0
    for spectral_parameter in spectral_parameters.values():
        if spectral_parameter.size > 3:
            dim += 1
        
    grid = {}
    if dim == 0:
        return grid
    elif dim == 1:
        n = 1000
    elif dim == 2:
        n = 100

    for key, value in spectral_parameters.items():
        if value.size > 3:
            grid_range = np.linspace(np.amin(value), np.amax(value), n)
            grid[key] = grid_range

    return grid


def interp1d(freqs, bandpass, grid, comp):
    """Performs 1D bandpass integration.
    
    This function uses numpy interp1d to perform bandpass integration
    for a sky component and returns the frequency scaling factor.
    
    Parameters:
    -----------
    freqs: `astropy.units.Quantity`
        Frequencies corresponding to the bandpass profile.
    bandpass: `astropy.units.Quantity`
        Normalized bandpass profile in units of 1/Hz.    
    grid: `astropy.units.Quantity`
        Linearly spaced grid points of the spectral parameters for which to
        integrate over.
    comp: `cosmoglobe.sky.Component`
        A Cosmoglobe sky component object.

    Returns:
    --------
    scaling: `astropy.units.Quantity`
        Frequency scaling factor obtained by integrating over the bandpass.
    """

    # Grid dictionary only has one item
    (key, grid), = grid.items()

    if comp._is_polarized:
        integrals = np.zeros((len(grid), 3))
    else:
        integrals = np.zeros((len(grid), 1))

    for idx, grid_point in enumerate(grid):
        freq_scaling = comp._get_freq_scaling(
            freqs, comp.freq_ref, **{key : grid_point}
        )
        integrals[idx] = np.trapz(freq_scaling*bandpass, freqs)

    # We transpose the array to make it into row format similar to how 
    # regular IQU maps are stored
    integrals = np.transpose(integrals)

    scaling =  [
        np.interp(comp.spectral_parameters[key][idx].value, grid.value, col)
        for idx, col in enumerate(integrals)
    ]

    return scaling


def interp2d(freqs, bandpass, grid, comp):
    """Performs 1D bandpass integration.
    
    This function uses scipy's RectBivariateSpline to perform  bandpass 
    integration for a sky component and returns the frequency scaling factor.

    TODO: Test this feature with a real chain.
    
    Parameters
    ----------
    freqs: `astropy.units.Quantity`
        Frequencies corresponding to the bandpass profile.
    bandpass: `astropy.units.Quantity`
        Normalized bandpass profile in units of 1/Hz.    
    grid: `astropy.units.Quantity`
        Linearly spaced 2D grid of the spectral parameters for which to
        integrate over.
    comp: `cosmoglobe.sky.Component`
        A Cosmoglobe sky component object.

    Returns
    -------
    scaling: `astropy.units.Quantity`
        Frequency scaling factor obtained by integrating over the bandpass.
    """

    # Make n x n mesh grid for the spectral parameters
    n = len(list(grid.values())[0])
    if comp._is_polarized:
        integrals = np.zeros((n, n, 3))
    else:
        integrals = np.zeros((n, n, 1))

    mesh_grid = {
        key: value for key, value in zip(
            grid.keys(), np.meshgrid(*grid.values())
        )
    }

    for i in range(n):
        for j in range(n):
            grid_spectrals = {key: value[i,j] for key, value in mesh_grid.items()}
            freq_scaling = comp._get_freq_scaling(
                freqs, comp.freq_ref, **grid_spectrals
            )
            integrals[i,j] = np.trapz(freq_scaling*bandpass, freqs)
    integrals = np.transpose(integrals)

    scaling = []
    for idx, col in enumerate(integrals):
        f = RectBivariateSpline(*grid.values(), col)
        packed_spectrals = [
            spectral[idx] if spectral.shape[0] == 3 else spectral[0]
            for spectral in comp.spectral_parameters.values()
        ]
        scaling.append(f(*packed_spectrals, grid=False))
    
    return scaling


u.quantity_input(freq=u.Hz)
def b_rj(freq):
    """Returns the intensity derivative for K_RJ.

    Computes the intensity derivate factor [Jy/sr -> K_RJ].

    Parameters
    ----------
    freq : `astropy.units.Quantity`
        Frequency [Hz].   

    Returns
    -------
    `astropy.units.Quantity`
        Intensity derivative factor.
    """

    return (2*const.k_B*freq**2) / (const.c**2 * u.sr)


u.quantity_input(freq=u.Hz, T=u.K)
def b_cmb(freq, T=const.T_0):
    """The intensity derivative for K_CMB.

    Computes the intensity derivate factor [Jy/sr -> K_CMB].

    Parameters
    ----------
    freq : `astropy.units.Quantity`
        Frequency [Hz].   
    T : `astropy.units.Quantity`
        The CMB emperature [K]. Default is const.T_0.

    Returns
    -------
    `astropy.units.Quantity`:
        Bandpass coefficient.
    """

    x = (const.h*freq) / (const.k_B*T)

    term1 = (2*const.h*freq**3) / (const.c**2 * np.expm1(x))
    term2 = np.exp(x) / np.expm1(x)
    term3 =  (const.h*freq) / (const.k_B*T**2)

    return term1 * term2 * term3 / u.sr


u.quantity_input(freq=u.Hz, freq_ref=u.Hz)
def b_iras(freq, freq_ref):
    """The intensity derivative in the IRAS unit convention (MJy/sr).

    Parameters
    ----------
    freq : `astropy.units.Quantity`
        Frequency [Hz].    
    freq_ref : `astropy.units.Quantity`
        Reference frequency [Hz].

    Returns
    -------
    `astropy.units.Quantity`
        Bandpass coefficient.
    """

    return freq_ref / freq


from enum import Enum, auto

import astropy.units as u
import numpy as np

import cosmoglobe.utils.functions as F

class CompState(Enum):
    """State of a model component."""

    ENABLED = auto()
    DISABLED = auto()


class NsideError(Exception):
    """Raised if there is a NSIDE related problem."""


class ModelError(Exception):
    """Raised if there is a Model related problem."""


def extract_scalars(spectral_parameters):
    """Extracts scalars from a spectral_parameter dict.
    
    If a map is constant in all dims, then it is converted to a (ndim,) 
    array.
    """

    scalars = {}
    for key, value in spectral_parameters.items():
        if not isinstance(value, u.Quantity):
            value = u.Quantity(value, unit=u.dimensionless_unscaled)
        if value.size > 1:
            if value.ndim > 1:
                uniques = [np.unique(col) for col in value.value]
                all_is_unique = all([len(col) == 1 for col in uniques])
            else:
                uniques = np.unique(value.value)
                all_is_unique = len(uniques) == 1
            if all_is_unique:
                scalars[key] = uniques*value.unit
        else:
            scalars[key] = value

    return scalars


def set_spectral_units(maps):
    """Sets the spectral value for a specific spectral parameter name."""

    #TODO: Figure out how to correctly detect unit of spectral map in chain.
    #      Until then, a hardcoded dict is used:
    units = dict(
        T=u.K,
        Te=u.K,
        nu_p=u.GHz,
    )
    for map_ in maps:
        if map_ in units:
            maps[map_] *= units[map_]
        else:
            maps[map_] *= u.dimensionless_unscaled

    return maps


def str_to_astropy_unit(unit):
    """Converts a K_RJ or a K_CMB string to `u.K` unit.
    
    This function preserves the prefix.
    """

    try:
        output_unit = u.Unit(unit)
    except ValueError:
        if unit.lower().endswith('k_rj'):
            output_unit = u.Unit(unit[:-3])
            
        elif unit.lower().endswith('k_cmb'):
            output_unit = u.Unit(unit[:-4])

    return output_unit


def emission_to_unit(emission, freqs, unit):
    """Converts the unit of the emission.
    
    Parameters
    ----------
    emission : `astropy.quantity.Quantity`
        Emission whos unit we want to convert.
    freqs : `astropy.quantity.Quantity`
        Frequency or a list of frequencies for which the emission was 
        obtained.
    unit : `astropy.UnitBase`
        Unit to convert to.
    """

    try:
        unit = u.Unit(unit)
        emission = emission.to(
            unit, equivalencies=u.brightness_temperature(freqs)
        )

    except ValueError:
        if unit.lower().endswith('k_rj'):
            unit = u.Unit(unit[:-3])
        elif unit.lower().endswith('k_cmb'):
            unit = u.Unit(unit[:-4])  
            emission *= F.brightness_to_thermodynamical(freqs)
        emission.to(unit)

    return emission


def gaussian_beam_2D(r, sigma):
    """Returns the Gaussian beam in 2D in polar coordinates.

    Parameters
    ----------
    r : float
        Angular distance.
    sigma : float
        The sigma of the Gaussian (beam radius).

    Returns
    -------
    `numpy.array_like`
        Gaussian beam.
    """

    return r*np.exp(-(r**2)/(2 * sigma**2))/(sigma*np.sqrt(2*np.pi))


import click

"""
THIS IS ALL CURRENTLY PSEUDOCODE
"""

@commands_plotting.command()
@click.argument("filename", type=click.STRING)
def plotfits(filename):
    """
    This function wraps around the plotting function and plots fits files
    input: filename of fits file, optional plotting params
    """
    from cosmoglobe.plot.mollweide import mollplot
    import fitsio
    data = fitsio.open(filename)
    map_ = filename.to_map()
    mollplot(map_)

@commands_plotting.command()
@click.argument("filename", type=click.STRING)
def plothdf(filename, dataset):
    """
    This function wraps around the plotting function and plots hdf files
    Input: hdf file, dataset and nside, optional plotting params
    """
    from cosmoglobe.plot.mollweide import mollplot
    import h5
    data = h5.open(filename, dataset)
    map_ = filename.to_map()
    mollplot(map_)



from abc import ABC, abstractmethod
from typing import Dict
from sys import exit
import sys
import warnings

import astropy.units as u
import healpy as hp
import numpy as np
from tqdm import tqdm

from cosmoglobe.utils import utils
import cosmoglobe.utils.bandpass as bp

warnings.simplefilter('once', UserWarning)


class _Component(ABC):
    """Abstract base class for a sky component"""

    amp: u.Quantity
    freq_ref: u.Quantity
    spectral_parameters: Dict[str, u.Quantity]

    def __init__(self, amp, freq_ref, **spectral_parameters):
        """Initializes a sky component."""
        
        self.amp = self._reshape_amp(amp)
        self.freq_ref = self._reshape_freq_ref(freq_ref)
        self.spectral_parameters = self._reshape_spectral_parameters(
            spectral_parameters
        )

    @u.quantity_input(
        freqs=u.Hz, 
        bandpass=(u.Jy/u.sr, u.K, None), 
        fwhm=(u.rad, u.deg, u.arcmin),
    )
    def __call__(
        self, freqs, bandpass=None, fwhm=0.0 * u.rad, output_unit=u.uK
    ):
        r"""Simulates the component sky emission. 

        This method computes the full component sky emission for a single 
        frequency :math:`\nu` or integrated over a  bandpass :math:`\tau`.

        Parameters
        ----------
        freqs : `astropy.units.Quantity`
            A frequency, or a list of frequencies for which to evaluate 
            the sky emission.
        bandpass : `astropy.units.Quantity`, optional
            Bandpass profile corresponding to the frequencies. Default is 
            None. If `bandpass` is None and `freqs` is a single frequency,
            a delta peak is assumed (unless `freqs` is a list of 
            frequencies, for which a top-hat bandpass is used to perform  
            bandpass integration instead).
        fwhm : `astropy.units.Quantity`, optional
            The full width half max parameter of the Gaussian (Default is 
            0.0, which indicates no smoothing of output maps).
        output_unit : str, `astropy.units.UnitBase`, optional
            The desired output units of the emission. The supported units are
            :math:`\mathrm{\mu K_{RJ}}` and :math:`\mathrm{MJ/sr}` (By 
            default the output unit of the model is always in 
            :math:`\mathrm{\mu K_{RJ}}`. 

        Returns
        -------
        `astropy.units.Quantity`
            The component emission.

        Notes
        -----
        This function computes the following expression for a given 
        component:
        
        .. math::

            \boldsymbol{s}_i(\nu) = \boldsymbol{a}
            _i(\nu_{0,i}) \; f_{i}
            (\nu),

        where :math:`\boldsymbol{a}_i` is the amplitude map of 
        component :math:`i` at some reference frequency 
        :math:`\nu_{0,i}`, and :math:`f_i(\nu)` is the 
        scaling factor given by the SED of the component.

        See Also
        --------
        Model.__call__
            See for explicit expressions for :math:`\boldsymbol{s}_i(\nu)`.

        Examples
        --------
        Simulated dust emission at :math:`350\; \mathrm{GHz}`:

        >>> from cosmoglobe import skymodel
        >>> import astropy.units as u
        >>> model = skymodel(nside=256) 
        >>> model.dust(350*u.GHz)[0]
        [21.12408523 -0.30044442  5.8907252  ...  4.6465226   6.94981578
          8.23490773] uK

        Simulated synchrotron emission at :math:`500\; \mathrm{GHz}` 
        smoothed with a :math:`50\; '` Gaussian beam, outputed in units of 
        :math:`\mathrm{MJy} / \mathrm{sr}`:

        >>> model.synch(40*u.GHz, fwhm=50*u.arcmin, output_unit='MJy/sr')[0]
        [0.00054551 0.00053428 0.00051471 ... 0.00046559 0.00047185 0.00047065]
         MJy / sr
        """

        if freqs.size == 1:
            emission = self._get_delta_emission(freqs)
            emission = self._smooth_emission(emission, fwhm)
            if output_unit is not None:
                emission = utils.emission_to_unit(emission, freqs, output_unit)
        else:
            emission = self._get_bandpass_emission(
                freqs, bandpass, output_unit=output_unit
            )
            emission = self._smooth_emission(emission, fwhm)

        return emission

    def _get_delta_emission(self, freq):
        """Simulates the component emission at a delta frequency.
        
        Parameters
        ----------
        freq : `astropy.units.Quantity`
            A delta frequency.

        Returns
        -------
        `astropy.units.Quantity`
            Simulated emission.
        """

        scaling = self._get_freq_scaling(
            freq, self.freq_ref, **self.spectral_parameters
        )

        return self.amp * scaling

    def _get_bandpass_emission(self, freqs, bandpass=None, output_unit=u.uK):
        """Computes the simulated component emission over a bandpass.

        If no bandpass is passed, a top-hat bandpass is assumed.

        Parameters
        ----------
        freqs : `astropy.units.Quantity`
            Bandpass frequencies.
        bandpass : `astropy.units.Quantity`
            Bandpass profile. Default: None
        output_unit : `astropy.units.UnitBase`
            The desired output unit of the emission. Default: None

        Returns
        -------
        emission : `astropy.units.Quantity`
            Simulated emission.
        """

        if bandpass is None:
            warnings.warn('Bandpass is None. Defaulting to top-hat bandpass')
            bandpass = np.ones(freqs_len := len(freqs))/freqs_len * u.K

        bandpass = bp.get_normalized_bandpass(bandpass, freqs)
        bandpass_coefficient = bp.get_bandpass_coefficient(
            bandpass, freqs, output_unit
        )

        bandpass_scaling = bp.get_bandpass_scaling(freqs, bandpass, self)
        emission = self.amp * bandpass_scaling * bandpass_coefficient
        
        return emission

    @property
    def _is_polarized(self):
        """Returns True if component is polarized and False if not."""

        if self.amp.shape[0] == 3:
            return True

        return False

    @staticmethod
    def _reshape_amp(amp):
        """Reshapes the reference frequency to a broadcastable shape."""

        return amp if amp.ndim != 1 else np.expand_dims(amp, axis=0)

    @staticmethod
    def _reshape_freq_ref(freq_ref):
        """Reshapes the reference frequency to a broadcastable shape."""

        if freq_ref is None:
            return
        elif freq_ref.size == 1:
            return freq_ref
        elif freq_ref.size == 2:
            return np.expand_dims(
                u.Quantity([freq_ref[0], freq_ref[1], freq_ref[1]]), axis=1
            )
        elif freq_ref.size == 3:
            return freq_ref.reshape((3,1))
        else:
            raise ValueError('Unrecognized shape.')

    @staticmethod
    def _reshape_spectral_parameters(spectral_parameters):
        """Reshapes the spectral parameters to a broadcastable shape."""

        return {
            key: (np.expand_dims(value, axis=0) if value.ndim == 1 else value)
            for key, value in spectral_parameters.items()
        }

    @abstractmethod
    def _smooth_emission(self, emission, fwhm):
        """Smooths simulated emission given a beam FWHM.

        Parameters
        ----------
        emission : `astropy.units.Quantity`
            Simulated emission.
        fwhm : `astropy.units.Quantity`
            The full width half max parameter of the Gaussian. Default: 0.0

        Returns
        -------
        emission : `astropy.units.Quantity`
            Smoothed emission.
        """

    def __repr__(self):
        """Representation of the component."""

        main_repr = f'{self.__class__.__name__}'
        main_repr += '('
        extra_repr = ''
        for key in self.spectral_parameters.keys():
            extra_repr += f'{key}, '
        if extra_repr:
            extra_repr = extra_repr[:-2]
        main_repr += extra_repr
        main_repr += ')'

        return main_repr


class _DiffuseComponent(_Component):
    """Abstract base class for a diffuse sky component."""

    def __init__(self, amp, freq_ref, **spectral_parameters):
        super().__init__(amp, freq_ref, **spectral_parameters)

    @abstractmethod
    def _get_freq_scaling(freq, freq_ref, **spectral_parameters):
        r"""Computes the frequency scaling for a component.
        
        Parameters
        ----------
        freq : `astropy.units.Quantity`
            Frequency at which to evaluate the model.
        freq_ref : `astropy.units.Quantity`
            Reference frequencies for the components `amp` map.
        spectral_parameters : `np.ndarray`, `astropy.units.Quantity`
            The spectral parameters of the component. 
            
        Returns
        -------
        scaling : `astropy.units.Quantity`
            Frequency scaling factor with dimensionless units.
        """

    def _smooth_emission(self, emission, fwhm):
        """See base class."""
        
        if fwhm == 0.0:
            return emission
        
        fwhm = (fwhm.to(u.rad)).value
        if self._is_polarized:
            emission = u.Quantity(
                hp.smoothing(emission, fwhm=fwhm), unit=emission.unit
            )
        else:
            emission[0] = u.Quantity(
                hp.smoothing(emission[0], fwhm=fwhm), unit=emission.unit
            )

        return emission


class _PointSourceComponent(_Component):
    """Abstract base class for a point source sky component."""

    def __init__(self, amp, freq_ref, nside, **spectral_parameters):
        super().__init__(amp, freq_ref, **spectral_parameters)

        # Point source components explicitly require a nside attribute 
        # since the amplitudes are not stored in healpix maps.
        self.nside = nside

    @abstractmethod
    def _get_freq_scaling(freq, freq_ref, **spectral_parameters):
        r"""Computes the frequency scaling for a component.
        
        Parameters
        ----------
        freq : `astropy.units.Quantity`
            Frequency at which to evaluate the model.
        freq_ref : `astropy.units.Quantity`
            Reference frequencies for the components `amp` map.
        spectral_parameters : `np.ndarray`, `astropy.units.Quantity`
            The spectral parameters of the component. 
            
        Returns
        -------
        scaling : `astropy.units.Quantity`
            Frequency scaling factor with dimensionless units.
        """

    def _smooth_emission(self, emission, fwhm):
        """See base class. 
        
        In addition to smoothing the point sources, this method also maps 
        the cataloged points to a healpix map in the process.
        """

        return self._points_to_map(emission, fwhm=fwhm)

    def _points_to_map(self, amp, fwhm, n_fwhm=2, verbose=False):
        """Maps the cataloged point sources onto a healpix map with a truncated 
        gaussian beam. For more information, see 
         `Mitra et al. (2010) <https://arxiv.org/pdf/1005.1929.pdf>`_.

        Parameters
        ----------
        amp : `astropy.units.Quantity`
            Amplitudes of the point sources.
        fwhm : `astropy.units.Quantity`
            The full width half max parameter of the Gaussian. Default: 0.0
        n_fwhm : int, float
            The fwhm multiplier used in computing radial cut off r_max 
            calculated as r_max = n_fwhm * fwhm of the Gaussian.
            Default: 2.
        nside : int
            The healpix map resolution of the output map. If component is part 
            of a sky model, we automatically select the model nside. 
            Default: None.
        """

        nside = self.nside
        angular_coords = self.angular_coords
        
        if amp.ndim > 1:
            amp = np.squeeze(amp)
        healpix_map = u.Quantity(
            np.zeros(hp.nside2npix(nside)), unit=amp.unit
        )
        pix_lon, pix_lat = hp.pix2ang(
            nside, np.arange(hp.nside2npix(nside)), lonlat=True
        )

        fwhm = fwhm.to(u.rad)
        sigma = fwhm / (2*np.sqrt(2*np.log(2)))

        if sigma == 0.0:
            # No smoothing nesecarry. We directly map the sources to pixels
            warnings.warn('mapping point sources to pixels without beam smoothing.')
            pixels = hp.ang2pix(nside, *angular_coords.T, lonlat=True)
            beam_area = hp.nside2pixarea(nside) * u.sr
            healpix_map[pixels] = amp
            healpix_map /= beam_area.value

        else:
            # Apply a truncated gaussian beam to each point source
            pix_res = hp.nside2resol(nside)
            if fwhm.value < pix_res:
                raise ValueError(
                    'fwhm must be >= pixel resolution to resolve the '
                    'point sources.'
                )
            beam_area = 2 * np.pi * sigma**2
            r_max = n_fwhm * fwhm.value

            with tqdm(
                total=len(angular_coords), file=sys.stdout, disable=not verbose
            ) as pbar:
                sigma = sigma.value
                beam = utils.gaussian_beam_2D
                print('Smoothing point sources...')

                for idx, (lon, lat) in enumerate(angular_coords):
                    vec = hp.ang2vec(lon, lat, lonlat=True)
                    inds = hp.query_disc(nside, vec, r_max)
                    r = hp.rotator.angdist(
                        np.array(
                            [pix_lon[inds], pix_lat[inds]]), np.array([lon, lat]
                        ),
                        lonlat=True
                    )
                    healpix_map[inds] += amp[idx] * beam(r, sigma)
                    pbar.update()

        healpix_map /= beam_area.value
        return np.expand_dims(healpix_map, axis=0)

    def _read_coords_from_catalog(self, catalog):
        """Reads in the angular coordinates of the point sources from a 
        given catalog.

        Parameters
        ----------
        catalog : str
            Path to the point source catalog. Default is the COM_GB6 
            catalog.
        
        Returns
        -------
        coords : `numpy.ndarray`
            Longitude and latitude values of each point source
        """
        
        try:
            coords = np.loadtxt(catalog, usecols=(0,1))
        except OSError:
            raise OSError('Could not find point source catalog')

        if len(coords) != len(self.amp[0]):
            raise ValueError('Cataloge does not match chain catalog')

        return coords


from pathlib import Path
from sys import exit
import warnings

import astropy.units as u
import numpy as np
import healpy as hp

from cosmoglobe.sky.base import (
    _DiffuseComponent,
    _PointSourceComponent,
)
from cosmoglobe.utils import functions as F


DATA_DIR = Path(__file__).parent.parent.resolve() / 'data'
RADIO_CATALOG = DATA_DIR / 'radio_catalog.dat'
SPDUST2_FILE = DATA_DIR / 'spdust2_cnm.dat'


class Synchrotron(_DiffuseComponent):
    r"""Class representing the synchrotron component.

    Attributes
    ----------
    amp : `astropy.units.Quantity`
        Emission maps of synchrotron at the reference frequencies given
        by `freq_ref` from a commander3 chain.
    freq_ref : `astropy.units.Quantity`
        Reference frequencies :math:`\nu_\mathrm{0,s}` for `amp`.
    spectral_parameters : dict
        Dictionary containing the power law spectral index :math:`\beta`.
    label : str
        Component label.

    Methods
    -------
    __call__

    Notes
    -----
    This is a generic power law given at a reference frequency 
    :math:`\nu_{s,0}` with a power law :math:`\beta` in Rayleigh-Jeans 
    temperature. It is defined using the convention in 
    `BeyondPlanck (2020), Section 3.3.1 
    <https://arxiv.org/pdf/2011.05609.pdf>`_;

    .. math::

        \boldsymbol{s}_\mathrm{RJ}^{\mathrm{s}}(\nu) \propto
        \left( \frac{\nu}{\nu_\mathrm{0,s}} \right)^
        {\beta + C \ln \nu / \nu_{0,s}},

    where :math:`\nu` is the frequency for which we are simulating the 
    sky emission. :math:`C` is set to 0 for all current implementations 
    as of BP9.
    """

    label = 'synch'

    def __init__(self, amp, freq_ref, beta):
        super().__init__(amp, freq_ref, beta=beta)

    def _get_freq_scaling(self, freq, freq_ref, beta):
        r"""See base class.

        Parameters
        ----------
        beta : `numpy.ndarray`, `astropy.units.Quantity`
            The power law spectral index.
        """

        scaling = (freq/freq_ref)**beta

        return scaling


class Dust(_DiffuseComponent):
    r"""Class representing the thermal dust component.
    
    Attributes
    ----------
    amp : `astropy.units.Quantity`
        Emission maps of thermal dust at the reference frequencies given
        by `freq_ref`.
    freq_ref : `astropy.units.Quantity`
        Reference frequencies :math:`\nu_\mathrm{0,d}` for `amp`.
    spectral_parameters : dict
        Dictionary containing the spectral parameters :math:`\beta` and 
        :math:`T`.
    label : str
        Component label.

    Methods
    -------
    __call__

    Notes
    -----
    This is a modified blackbody with a power law spectral index 
    :math:`\beta_\mathrm{d}` in Rayleigh-Jeans temperature, and thermal dust 
    temperature :math:`T_{\mathrm{d}}`. It is defined using the convention 
    in `BeyondPlanck (2020), Section 3.3.3 
    <https://arxiv.org/pdf/2011.05609.pdf>`_;

    .. math::

        \boldsymbol{s}_\mathrm{RJ}^{\mathrm{d}}(\nu) \propto 
        \frac{\nu^{\beta_{\mathrm{d}}+1}}{\mathrm{e}^
        {h\nu/kT_{\mathrm{d}}}-1},
    
    where :math:`\nu` is the frequency for which we are simulating the 
    sky emission, :math:`h` is Planck's constant, and :math:`k` is the 
    Boltzmann constant.
    """

    label = 'dust'

    def __init__(self, amp, freq_ref, beta, T):
        super().__init__(amp, freq_ref, beta=beta, T=T)

    def _get_freq_scaling(self, freq, freq_ref, beta, T):
        r"""See base class.

        Parameters
        ----------
        beta : `numpy.ndarray`, `astropy.units.Quantity`
            The power law spectral index.
        T : `astropy.units.Quantity`
            Temperature of the blackbody.
        """

        blackbody_ratio = (
            F.blackbody_emission(freq, T) / F.blackbody_emission(freq_ref, T)
        )
        scaling = (freq/freq_ref)**(beta-2) * blackbody_ratio

        return scaling


class FreeFree(_DiffuseComponent):
    r"""Class representing the free-free component.

    Attributes
    ----------
    amp : `astropy.units.Quantity`
        Emission maps of free-free at the reference frequencies given
        by `freq_ref`.
    freq_ref : `astropy.units.Quantity`
        Reference frequencies :math:`\nu_\mathrm{0,\mathrm{ff}}` for `amp`.
    spectral_parameters : dict
        Dictionary containing the spectral parameter :math:`T_e`.
    label : str
        Component label.

    Methods
    -------
    __call__

    Notes
    -----
    The free-free emission is defined using the convention in 
    `BeyondPlanck (2020), Section 3.3.2 
    <https://arxiv.org/pdf/2011.05609.pdf>`_;

    .. math::

        \boldsymbol{s}_\mathrm{RJ}^{\mathrm{ff}}(\nu) \propto
        \frac{g_{\mathrm{ff}}(T_\mathrm{e})}{\nu^2},

    where :math:`\nu` is the frequency for which we are simulating the 
    sky emission, :math:`g_\mathrm{ff}` is the Gaunt factor, and 
    :math:`T_\mathrm{e}` is the electron temperature.
    """

    label = 'ff'

    def __init__(self, amp, freq_ref, Te):
        super().__init__(amp, freq_ref, Te=Te)

    def _get_freq_scaling(self, freq, freq_ref, Te):
        r"""See base class.
        
        Parameters
        ----------
        Te : `astropy.units.Quantity`
            Electron temperature.
        """

        gaunt_factor_ratio = (
            F.gaunt_factor(freq, Te) / F.gaunt_factor(freq_ref, Te)
        )
        scaling = (freq_ref/freq)**2 * gaunt_factor_ratio

        return scaling


class AME(_DiffuseComponent):
    r"""Class representing the spinning dust component.

    Attributes
    ----------
    amp : `astropy.units.Quantity`
        Emission maps of spinning dust at the reference frequencies given
        by `freq_ref`.
    freq_ref : `astropy.units.Quantity`
        Reference frequencies :math:`\nu_\mathrm{0,sd}` for `amps`.
    spectral_parameters : dict
        Dictionary containing the spectral parameter :math:`\nu_p`.
    label : str
        Component label.
    spdust2 : `astropy.units.Quantity`
        spdust2 template.

    Methods
    -------
    __call__

    Notes
    -----
    The spinning dust emission is defined using the convention in 
    `BeyondPlanck (2020), Section 3.3.4 
    <https://arxiv.org/pdf/2011.05609.pdf>`_;

    .. math::

        \boldsymbol{s}_{\mathrm{RJ}}^{\mathrm{sd}}(\nu) \propto 
        \nu^{-2} \boldsymbol{s}_{0}^{\mathrm{sd}}\left(\nu \cdot 
        \frac{30.0\; \mathrm{GHz}}{\nu_{p}}\right)

    where :math:`\nu` is the frequency for which we are simulating the 
    sky emission, :math:`\boldsymbol{s}_0^{\mathrm{sd}}` is the `spdust2` 
    template, and :math:`\nu_p` the peak frequency.
    """

    label = 'ame'

    def __init__(self, amp, freq_ref, nu_p):
        spdust2_freq, spdust2_amp = np.loadtxt(SPDUST2_FILE, unpack=True)
        spdust2_freq = u.Quantity(spdust2_freq, unit=u.GHz)
        spdust2_amp = u.Quantity(spdust2_amp, unit=(u.Jy/u.sr))
        spdust2_amp = spdust2_amp.to(
            u.K, equivalencies=u.brightness_temperature(spdust2_freq)
        )        
        self.spdust2 = np.array([spdust2_freq.si.value, spdust2_amp.si.value])

        super().__init__(amp, freq_ref, nu_p=nu_p)


    def _get_freq_scaling(self, freq, freq_ref, nu_p):
        r"""See base class.
        
        Parameters
        ----------
        nu_p : `astropy.units.Quantity`
            Peak frequency.
        """

        spdust2 = self.spdust2
        peak_scale = 30*u.GHz / nu_p

        # AME is undefined at outside of this frequency range
        if not np.logical_and(
            (freq*peak_scale).si.value > np.min(spdust2[0]),
            (freq*peak_scale).si.value < np.max(spdust2[0])
        ).all():
            return u.Quantity(0, unit=u.dimensionless_unscaled)

        interp = np.interp((freq * peak_scale).si.value, spdust2[0], spdust2[1])
        interp_ref = (
            np.interp((freq_ref * peak_scale).si.value, spdust2[0], spdust2[1])
        )
        scaling = interp / interp_ref

        return scaling


class CMB(_DiffuseComponent):
    r"""Class representing the CMB component.

    Attributes
    ----------
    amp : `astropy.units.Quantity`
        Emission map of CMB in units of :math:`\mathrm{\mu K_{CMB}}`
    freq_ref : `astropy.units.Quantity`
        Reference frequency is None due to the amplitude maps being 
        stored in units of :math:`\mathrm{\mu K_{CMB}}`.
    spectral_parameters : dict
        Empty dictionary.
    label : str
        Component label.

    Methods
    -------
    __call__
    remove_dipole
    get_dipole

    Notes
    -----
    The CMB emission is defined using the convention in 
    `BeyondPlanck (2020), Section 3.2 
    <https://arxiv.org/pdf/2011.05609.pdf>`_;

    .. math::

        \boldsymbol{s}_{\mathrm{RJ}}^{\mathrm{CMB}}(\nu) \propto 
        \frac{x^{2} \mathrm{e}^{x}}{\left(\mathrm{e}^{x}-1\right)
        ^{2}} \boldsymbol{s}^{\mathrm{CMB}},

    where :math:`\nu` is the frequency for which we are simulating the 
    sky emission, :math:`x=h \nu / k T_{0}` and 
    :math:`T_0 = 2.7255 \mathrm{K}` as of BP9.
    """

    label = 'cmb'

    def __init__(self, amp, freq_ref=None):
        super().__init__(amp, freq_ref=freq_ref)

    @u.quantity_input(gal_cut=u.deg)
    def get_dipole(self, gal_cut=10 * u.deg):
        """Returns the solar dipole from the reference amplitude map.

        Parameters
        ----------
        gal_cut : float, optional
            Masks pixles :math:`\pm` `gal_cut` in latitude before estimating 
            dipole (:math:`10^\circ` by default).

        Returns
        -------
        `astropy.units.Quantity`
            Map of the solar dipole.
        """

        if hasattr(self, 'dipole'):
            warnings.warn(
                'Returning previously removed dipole signal '
                '(overriding gal_cut).'
            )
            return self.dipole

        amp_without_dipole = u.Quantity(
                hp.remove_dipole(self.amp[0], gal_cut=gal_cut.value),
                unit=self.amp.unit
        )
        
        return self.amp[0] - amp_without_dipole

    @u.quantity_input(gal_cut=u.deg)
    def remove_dipole(self, gal_cut=10 * u.deg):
        """Removes the solar dipole from the reference amplitude map.

        Parameters
        ----------
        gal_cut : float, optional
            Masks pixles :math:`\pm` `gal_cut` in latitude before 
            estimating dipole (:math:`10^\circ` by default).
        """

        amp_without_dipole = u.Quantity(
                hp.remove_dipole(self.amp[0], gal_cut=gal_cut.value),
                unit=self.amp.unit
        )
        
        self.dipole = self.amp[0] - amp_without_dipole
        self.amp[0] = amp_without_dipole

    def _get_freq_scaling(self, freq, freq_ref=None):
        r"""See base class."""

        # We explicitly expand the dims to support broadcasting
        return np.expand_dims(
            F.thermodynamical_to_brightness(freq), axis=0
        )


class Radio(_PointSourceComponent):
    r"""Class representing the radio component.

    Attributes
    ----------
    amp : `astropy.units.Quantity`
        Point source amplitudes at the reference frequencies given
        by `freq_ref`. Note that this quantity is not a healpix map.
    freq_ref : `astropy.units.Quantity`
        Reference frequencies :math:`\nu_\mathrm{0,\mathrm{src}}` for 
        `amp`.
    spectral_parameters : dict
        Dictionary containing the spectral parameter :math:`\alpha`.
    nside : int
        Point source components need to be explicitly passed an NSIDE.
    label : str
        Component label.

    Methods
    -------
    __call__

    Notes
    -----
    This is a generic power law given at a reference frequency 
    :math:`\nu_{\mathrm{0, src}}` with a power law spectral index 
    :math:`\alpha`. It is defined using the convention in 
    `BeyondPlanck (2020), Section 3.4.1 
    <https://arxiv.org/pdf/2011.05609.pdf>`_;

    .. math::

        \boldsymbol{s}_{\mathrm{RJ}}^{\mathrm{src}}(\nu) \propto
        \left(\frac{\nu}{\nu_{\mathrm{0, src}}}\right)^{\alpha-2}

    where :math:`\nu` is the frequency for which we are simulating the 
    sky emission.
    """

    label = 'radio'

    def __init__(self, amp, freq_ref, nside, specind):
        super().__init__(amp, freq_ref, nside, specind=specind)

        self.angular_coords = self._read_coords_from_catalog(RADIO_CATALOG)

        # Note that the true unit of self.amp at this stage is "mJy". Here we 
        # manually set it to "mJy/sr" to make it convertable to uK. This does 
        # not cause any problems, since in all cases, we end up dividing the 
        # emission by some beam_area.
        self.amp = u.Quantity(self.amp.value, unit='mJy/sr')
        self.amp = self.amp.to(
            u.uK, equivalencies=u.thermodynamic_temperature(self.freq_ref)
        )

    def _get_freq_scaling(self, freq, freq_ref, specind):
        r"""See base class.

        Parameters
        ----------
        specind : `astropy.units.Quantity`
            Power law spectral index.
        """

        scaling = (freq/freq_ref)**(specind-2)

        return scaling